// This is the main DLL file.

#include "stdafx.h"
#include "nfapinet.h"
#include "nfapi.h"
#include "nfutil.h"

using namespace System::Text;
using namespace nfapinet;

// Helper routines
inline void copyArray(unsigned char * dst, cli::array<unsigned char> ^% src, int len)
{
	if (!src)
		return;

	pin_ptr<unsigned char> p = &src[0];
	memcpy(dst, p, min(len, src->Length));
}

inline void copyArray(cli::array<unsigned char> ^% dst, const unsigned char * src, int len)
{
	dst = gcnew cli::array<unsigned char>(len);		

	pin_ptr<unsigned char> p = &dst[0];
	memcpy(p, src, len);
}

inline void convertTcpConn(nfapinet::NF_TCP_CONN_INFO ^ dst, nfapi::PNF_TCP_CONN_INFO src)
{
	dst->filteringFlag = src->filteringFlag;
	dst->processId = src->processId;
	dst->direction = src->direction;
	dst->ip_family = src->ip_family;

	copyArray(dst->localAddress, src->localAddress, NF_MAX_ADDRESS_LENGTH);
	copyArray(dst->remoteAddress, src->remoteAddress, NF_MAX_ADDRESS_LENGTH);
}

inline void convertUdpConn(nfapinet::NF_UDP_CONN_INFO ^ dst, nfapi::PNF_UDP_CONN_INFO src)
{
	dst->processId = src->processId;
	dst->ip_family = src->ip_family;

	copyArray(dst->localAddress, src->localAddress, NF_MAX_ADDRESS_LENGTH);
}

inline void convertUdpConnReq(nfapinet::NF_UDP_CONN_REQUEST ^ dst, nfapi::PNF_UDP_CONN_REQUEST src)
{
	dst->filteringFlag = src->filteringFlag;
	dst->processId = src->processId;
	dst->ip_family = src->ip_family;

	copyArray(dst->localAddress, src->localAddress, NF_MAX_ADDRESS_LENGTH);
	copyArray(dst->remoteAddress, src->remoteAddress, NF_MAX_ADDRESS_LENGTH);
}

inline void stringToBuffer(System::String^ str, char * buf, int len)
{
	Encoding ^ e = gcnew ASCIIEncoding();
	cli::array<unsigned char> ^ s = e->GetBytes(str);
	int i;

	for (i=0; (i<s->Length) && (i<(len-1)); i++)
	{
		buf[i] = *((unsigned char ^)s->GetValue(i));
	}
	buf[i] = '\0';
}

// Managed event handler
ref class NFAPI_EH
{
public:
	static nfapinet::NF_EventHandler ^ m_pHandler;

	static void threadStart()
	{
		m_pHandler->threadStart();
	}
	
	static void threadEnd()
	{
		m_pHandler->threadEnd();
	}

	static void tcpConnectRequest(ENDPOINT_ID id, nfapi::PNF_TCP_CONN_INFO pConnInfo)
	{
		nfapinet::NF_TCP_CONN_INFO ^ connInfo = gcnew nfapinet::NF_TCP_CONN_INFO();
		convertTcpConn(connInfo, pConnInfo);
		
		m_pHandler->tcpConnectRequest(id, connInfo);

		pConnInfo->filteringFlag = connInfo->filteringFlag;
		copyArray(pConnInfo->remoteAddress, connInfo->remoteAddress, sizeof(pConnInfo->remoteAddress));
	}

	static void tcpConnected(ENDPOINT_ID id, nfapi::PNF_TCP_CONN_INFO pConnInfo)
	{
		nfapinet::NF_TCP_CONN_INFO ^ connInfo = gcnew nfapinet::NF_TCP_CONN_INFO();
		convertTcpConn(connInfo, pConnInfo);
		
		m_pHandler->tcpConnected(id, connInfo);
	}

	static void tcpClosed(ENDPOINT_ID id, nfapi::PNF_TCP_CONN_INFO pConnInfo)
	{
		nfapinet::NF_TCP_CONN_INFO ^ connInfo = gcnew nfapinet::NF_TCP_CONN_INFO();
		convertTcpConn(connInfo, pConnInfo);

		m_pHandler->tcpClosed(id, connInfo);
	}

	static void tcpReceive(ENDPOINT_ID id, const char * buf, int len)
	{
		m_pHandler->tcpReceive(id, (unsigned char*)buf, len);
	}

	static void tcpSend(ENDPOINT_ID id, const char * buf, int len)
	{
		m_pHandler->tcpSend(id, (unsigned char*)buf, len);
	}

	static void tcpCanReceive(ENDPOINT_ID id)
	{
		m_pHandler->tcpCanReceive(id);
	}

	static void tcpCanSend(ENDPOINT_ID id)
	{
		m_pHandler->tcpCanSend(id);
	}

	static void udpCreated(ENDPOINT_ID id, nfapi::PNF_UDP_CONN_INFO pConnInfo)
	{
		nfapinet::NF_UDP_CONN_INFO ^ connInfo = gcnew nfapinet::NF_UDP_CONN_INFO();
		convertUdpConn(connInfo, pConnInfo);

		m_pHandler->udpCreated(id, connInfo);
	}

	static void udpConnectRequest(ENDPOINT_ID id, nfapi::PNF_UDP_CONN_REQUEST pConnReq)
	{
		nfapinet::NF_UDP_CONN_REQUEST ^ connReq = gcnew nfapinet::NF_UDP_CONN_REQUEST();
		convertUdpConnReq(connReq, pConnReq);

		m_pHandler->udpConnectRequest(id, connReq);

		pConnReq->filteringFlag = connReq->filteringFlag;
		copyArray(pConnReq->remoteAddress, connReq->remoteAddress, sizeof(pConnReq->remoteAddress));
	}

	static void udpClosed(ENDPOINT_ID id, nfapi::PNF_UDP_CONN_INFO pConnInfo)
	{
		nfapinet::NF_UDP_CONN_INFO ^ connInfo = gcnew nfapinet::NF_UDP_CONN_INFO();
		convertUdpConn(connInfo, pConnInfo);

		m_pHandler->udpClosed(id, connInfo);
	}

	static void udpReceive(ENDPOINT_ID id, const unsigned char * remoteAddress, const char * buf, int len, nfapi::PNF_UDP_OPTIONS options)
	{
		cli::array<unsigned char> ^ remoteAddress_cs;
		NF_UDP_OPTIONS ^ options_cs = gcnew nfapinet::NF_UDP_OPTIONS();
		
		copyArray(remoteAddress_cs, remoteAddress, NF_MAX_ADDRESS_LENGTH);
		options_cs->flags = options->flags;
		if (options->optionsLength)
		{
			copyArray(options_cs->options, options->options, options->optionsLength);
		}

		m_pHandler->udpReceive(id, remoteAddress_cs, (unsigned char*)buf, len, options_cs);
	}

	static void udpSend(ENDPOINT_ID id, const unsigned char * remoteAddress, const char * buf, int len, nfapi::PNF_UDP_OPTIONS options)
	{
		cli::array<unsigned char> ^ remoteAddress_cs;
		NF_UDP_OPTIONS ^ options_cs = gcnew nfapinet::NF_UDP_OPTIONS();
		
		copyArray(remoteAddress_cs, remoteAddress, NF_MAX_ADDRESS_LENGTH);
		options_cs->flags = options->flags;
		if (options->optionsLength)
		{
			copyArray(options_cs->options, options->options, options->optionsLength);
		}

		m_pHandler->udpSend(id, remoteAddress_cs, (unsigned char*)buf, len, options_cs);
	}

	static void udpCanReceive(ENDPOINT_ID id)
	{
		m_pHandler->udpCanReceive(id);
	}

	static void udpCanSend(ENDPOINT_ID id)
	{
		m_pHandler->udpCanSend(id);
	}

};

// Native event handler
class EventHandlerInternal : public nfapi::NF_EventHandler
{
public:
	EventHandlerInternal()
	{
	}
	
	void threadStart()
	{
		NFAPI_EH::threadStart();
	}

	void threadEnd()
	{
		NFAPI_EH::threadEnd();
	}

	void tcpConnectRequest(ENDPOINT_ID id, nfapi::PNF_TCP_CONN_INFO pConnInfo)
	{
		NFAPI_EH::tcpConnectRequest(id, pConnInfo);
	}

	void tcpConnected(ENDPOINT_ID id, nfapi::PNF_TCP_CONN_INFO pConnInfo)
	{
		NFAPI_EH::tcpConnected(id, pConnInfo);
	}

	void tcpClosed(ENDPOINT_ID id, nfapi::PNF_TCP_CONN_INFO pConnInfo)
	{
		NFAPI_EH::tcpClosed(id, pConnInfo);
	}

	void tcpReceive(ENDPOINT_ID id, const char * buf, int len)
	{
		NFAPI_EH::tcpReceive(id, buf, len);
	}

	void tcpSend(ENDPOINT_ID id, const char * buf, int len)
	{
		NFAPI_EH::tcpSend(id, buf, len);
	}

	void tcpCanReceive(ENDPOINT_ID id)
	{
		NFAPI_EH::tcpCanReceive(id);
	}

	void tcpCanSend(ENDPOINT_ID id)
	{
		NFAPI_EH::tcpCanSend(id);
	}

	void udpCreated(ENDPOINT_ID id, nfapi::PNF_UDP_CONN_INFO pConnInfo)
	{
		NFAPI_EH::udpCreated(id, pConnInfo);
	}

	void udpConnectRequest(ENDPOINT_ID id, nfapi::PNF_UDP_CONN_REQUEST pConnReq)
	{
		NFAPI_EH::udpConnectRequest(id, pConnReq);
	}

	void udpClosed(ENDPOINT_ID id, nfapi::PNF_UDP_CONN_INFO pConnInfo)
	{
		NFAPI_EH::udpClosed(id, pConnInfo);
	}

	void udpReceive(ENDPOINT_ID id, const unsigned char * remoteAddress, const char * buf, int len, nfapi::PNF_UDP_OPTIONS options)
	{
		NFAPI_EH::udpReceive(id, remoteAddress, buf, len, options);
	}

	void udpSend(ENDPOINT_ID id, const unsigned char * remoteAddress, const char * buf, int len, nfapi::PNF_UDP_OPTIONS options)
	{
		NFAPI_EH::udpSend(id, remoteAddress, buf, len, options);
	}

	void udpCanReceive(ENDPOINT_ID id)
	{
		NFAPI_EH::udpCanReceive(id);
	}

	void udpCanSend(ENDPOINT_ID id)
	{
		NFAPI_EH::udpCanSend(id);
	}
};

// Native event handler
EventHandlerInternal m_eventHandler;


// NFAPI methods 
nfapinet::NF_STATUS NFAPI::nf_init(System::String^ driverName, nfapinet::NF_EventHandler ^ pHandler)
{
	char szDriverName[MAX_PATH];

	stringToBuffer(driverName, szDriverName, sizeof(szDriverName));

	NFAPI_EH::m_pHandler = pHandler;

	return (nfapinet::NF_STATUS)nfapi::nf_init(szDriverName, &m_eventHandler);
}

void NFAPI::nf_free()
{
	nfapi::nf_free();
}

nfapinet::NF_STATUS NFAPI::nf_registerDriver(System::String^ driverName)
{
	char szDriverName[MAX_PATH];

	stringToBuffer(driverName, szDriverName, sizeof(szDriverName));

	return (nfapinet::NF_STATUS)nfapi::nf_registerDriver(szDriverName);
}

nfapinet::NF_STATUS NFAPI::nf_unRegisterDriver(System::String^ driverName)
{
	char szDriverName[MAX_PATH];

	stringToBuffer(driverName, szDriverName, sizeof(szDriverName));

	return (nfapinet::NF_STATUS)nfapi::nf_unRegisterDriver(szDriverName);
}


nfapinet::NF_STATUS NFAPI::nf_tcpSetConnectionState(ENDPOINT_ID id, BOOL suspended)
{
	return (nfapinet::NF_STATUS)nfapi::nf_tcpSetConnectionState(id, suspended);
}

nfapinet::NF_STATUS NFAPI::nf_tcpPostSend(ENDPOINT_ID id, const unsigned char * buf, int len)
{
	return (nfapinet::NF_STATUS)nfapi::nf_tcpPostSend(id, (char*)buf, len);
}

nfapinet::NF_STATUS NFAPI::nf_tcpPostReceive(ENDPOINT_ID id, const unsigned char * buf, int len)
{
	return (nfapinet::NF_STATUS)nfapi::nf_tcpPostReceive(id, (char*)buf, len);
}

nfapinet::NF_STATUS NFAPI::nf_tcpClose(ENDPOINT_ID id)
{
	return (nfapinet::NF_STATUS)nfapi::nf_tcpClose(id);
}

nfapinet::NF_STATUS NFAPI::nf_udpSetConnectionState(ENDPOINT_ID id, BOOL suspended)
{
	return (nfapinet::NF_STATUS)nfapi::nf_udpSetConnectionState(id, suspended);
}

nfapinet::NF_STATUS NFAPI::nf_udpPostSend(ENDPOINT_ID id, 
								cli::array<unsigned char> ^ remoteAddress, 
								const unsigned char * buf, int len,
								nfapinet::NF_UDP_OPTIONS ^ options)
{
	unsigned char remoteAddressNative[NF_MAX_ADDRESS_LENGTH];
	nfapi::PNF_UDP_OPTIONS optionsNative = 
		(nfapi::PNF_UDP_OPTIONS)malloc(
			sizeof(nfapi::NF_UDP_OPTIONS)-1 + ((options->options)? options->options->Length : 0));

	if (!optionsNative)
		return (nfapinet::NF_STATUS)NF_STATUS_FAIL;

	copyArray(remoteAddressNative, remoteAddress, NF_MAX_ADDRESS_LENGTH);
	optionsNative->flags = options->flags;
	if (options->options && options->options->Length > 0)
	{
		optionsNative->optionsLength = options->options->Length;
		copyArray(optionsNative->options, options->options, options->options->Length);
	} else
	{
		optionsNative->optionsLength = 0;
	}

	nfapinet::NF_STATUS status = (nfapinet::NF_STATUS)nfapi::nf_udpPostSend(id, remoteAddressNative, (char*)buf, len, optionsNative);
	
	free(optionsNative);

	return status;
}

nfapinet::NF_STATUS NFAPI::nf_udpPostReceive(ENDPOINT_ID id, 
								   cli::array<unsigned char> ^ remoteAddress, 
								   const unsigned char * buf, int len,
								   nfapinet::NF_UDP_OPTIONS ^ options)
{
	unsigned char remoteAddressNative[NF_MAX_ADDRESS_LENGTH];
	nfapi::PNF_UDP_OPTIONS optionsNative = 
		(nfapi::PNF_UDP_OPTIONS)malloc(
			sizeof(nfapi::NF_UDP_OPTIONS)-1 + ((options->options)? options->options->Length : 0));

	if (!optionsNative)
		return (nfapinet::NF_STATUS)NF_STATUS_FAIL;

	copyArray(remoteAddressNative, remoteAddress, NF_MAX_ADDRESS_LENGTH);
	optionsNative->flags = options->flags;
	if (options->options && options->options->Length > 0)
	{
		optionsNative->optionsLength = options->options->Length;
		copyArray(optionsNative->options, options->options, options->options->Length);
	} else
	{
		optionsNative->optionsLength = 0;
	}

	nfapinet::NF_STATUS status = (nfapinet::NF_STATUS)nfapi::nf_udpPostReceive(id, remoteAddressNative, (char*)buf, len, optionsNative);

	free(optionsNative);

	return status;
}

inline void convertRule(nfapi::PNF_RULE dst, nfapinet::NF_RULE ^ src)
{
	memset(dst, 0, sizeof(*dst));

	dst->protocol = src->protocol;
	dst->processId = src->processId;
	dst->direction = src->direction;
	dst->localPort = src->localPort;
	dst->remotePort = src->remotePort;
	dst->ip_family = src->ip_family;
	dst->filteringFlag = src->filteringFlag;

	copyArray(dst->localIpAddress, src->localIpAddress, NF_MAX_IP_ADDRESS_LENGTH);
	copyArray(dst->localIpAddressMask, src->localIpAddressMask, NF_MAX_IP_ADDRESS_LENGTH);
	copyArray(dst->remoteIpAddress, src->remoteIpAddress, NF_MAX_IP_ADDRESS_LENGTH);
	copyArray(dst->remoteIpAddressMask, src->remoteIpAddressMask, NF_MAX_IP_ADDRESS_LENGTH);
}

nfapinet::NF_STATUS NFAPI::nf_addRule(nfapinet::NF_RULE ^ pRule, int toHead)
{
	nfapi::NF_RULE rule;

	convertRule(&rule, pRule);
	
	return (nfapinet::NF_STATUS) nfapi::nf_addRule(&rule, toHead);
}

nfapinet::NF_STATUS NFAPI::nf_deleteRules()
{
	return (nfapinet::NF_STATUS) nfapi::nf_deleteRules();
}

unsigned long NFAPI::nf_setTCPTimeout(unsigned long timeout)
{
	return nfapi::nf_setTCPTimeout(timeout);
}

nfapinet::NF_STATUS NFAPI::nf_tcpDisableFiltering(ENDPOINT_ID id)
{
	return (nfapinet::NF_STATUS)nfapi::nf_tcpDisableFiltering(id);
}

nfapinet::NF_STATUS NFAPI::nf_udpDisableFiltering(ENDPOINT_ID id)
{
	return (nfapinet::NF_STATUS)nfapi::nf_udpDisableFiltering(id);
}

nfapinet::NF_STATUS NFAPI::nf_disableFiltering()
{
	return (nfapinet::NF_STATUS)nfapi::nf_disableFiltering();
}

System::String^ NFAPI::nf_getProcessName(unsigned long processId)
{
	wchar_t buf[MAX_PATH];

	if (nfapi::nf_getProcessName(processId, buf, sizeof(buf)/sizeof(buf[0])))
	{
		return gcnew System::String(buf);
	}

	return gcnew System::String(L"System");
}
